"""Velocity backends."""

from re import Match as re_Match, sub as re_sub, match as re_match
from loguru import logger
from shutil import which as shutil_which
from pathlib import Path
from abc import ABC, abstractmethod
from ._exceptions import (
    RepeatedSection,
    LineOutsideOfSection,
    TemplateSyntaxError,
    BackendNotSupported,
    BackendNotAvailable,
)
from ._config import config
from ._graph import Image


def _substitute(text: str, variables: dict[str, str], regex: str) -> str:
    """Substitute a variables in a string by a regex."""

    def _replace(m: re_Match):
        try:
            return str(variables[m.group(1)])
        except KeyError:
            logger.warning("The variable '{}' is undefined. Setting value to ''.".format(m.group(1)))

    return re_sub(regex, _replace, text)


class Backend(ABC):
    """Abstract class for velocity backend."""

    template_sections: list[str] = [
        "@from",
        "@pre",
        "@copy",
        "@run",
        "@env",
        "@label",
        "@entry",
        "@post",
    ]

    @classmethod
    def _get_sections(cls, template: list[str]) -> dict[str, list[str]]:
        """Retrieve the sections from a VTMP."""

        sections: dict[str, list[str]] = dict()
        past_sections: list[str] = list()
        current_section: str | None = None

        for line in template:
            # if a line contains a section header
            if line in cls.template_sections:
                # move the current section to past_sections
                if current_section is not None:
                    past_sections.append(current_section)

                # check for repeated section
                if line in past_sections:
                    raise RepeatedSection(f"You have more than one '{line}' section in your template!")
                else:
                    current_section = line
                    if current_section not in sections:
                        sections[current_section] = list()

            else:
                # handel content lines
                if current_section is None:
                    raise LineOutsideOfSection("You have a line outside of a section in your template!")
                else:
                    sections[current_section].append(line)

        # remove empty sections
        del_list: list[str] = list()
        for sec in sections:
            if len(sections[sec]) == 0:
                del_list.append(sec)
        for sec in del_list:
            del sections[sec]

        return sections

    @classmethod
    def _filter_content(cls, image: Image, text: str) -> str:
        """Filter conditionals and white space from a template line."""
        # handle conditionals
        res: re_Match[str] = re_match(r".*(\?\?([\S ]*)\|>(.*)\?\?).*", text)
        if res is not None:
            if image.satisfies(res.group(2)):
                text = re_sub(r"(\?\?.*\?\?)", res.group(3).strip(), text)
            else:
                text = re_sub(r"(\?\?.*\?\?)", "", text)

        # remove comments, newlines, and superfluous white space
        text = re_sub(r">>>.*", "", text)
        text = re_sub(r"\n", "", text)
        text = re_sub(r"^\s+|\s+$", "", text)

        return text

    @classmethod
    def _load_template(cls, image: Image, variables: dict[str, str]) -> list[str]:
        """Load a template and parse it."""
        template: list[str] = list()
        with open(
            Path(image.path).joinpath("templates", "{}.vtmp".format(image.template)),
            "r",
        ) as file:
            contents: list[str] = file.readlines()
            for line in contents:
                fcon: str = cls._filter_content(image, _substitute(line, variables, r"{{\s*(\S+)\s*}}"))
                if fcon != "":
                    template.append(fcon)
        return template

    def __init__(self, name: str, executable: str) -> None:
        self.name: str = name
        self.executable: str = executable

    def generate_script(self, image: Image, variables: dict[str, str]) -> list[str]:
        """Generate a build script e.g. .dockerfile/.def"""
        logger.debug("Variables: {}".format(variables))
        template: list[str] = self._load_template(image, variables)
        sections: dict[str, list[str]] = self._get_sections(template)
        script: list[str] = list()

        # @from
        try:
            if len(sections["@from"]) != 1:
                raise TemplateSyntaxError(
                    "You can only have one source in the @from section!",
                )
            elif len(sections["@from"][0].split()) != 1:
                raise TemplateSyntaxError("Your source must be a single string!", sections["@from"][0])
            else:
                script.extend(self._from(sections["@from"]))
        except KeyError:
            raise TemplateSyntaxError("You must have an @from section in your template!")
        # arguments
        script.extend(self._arguments(sections))
        # @pre
        if "@pre" in sections:
            script.extend(self._literal_section(sections["@pre"]))
        # @copy
        if "@copy" in sections:
            script.extend(self._copy(sections["@copy"]))
        # @run
        if "@run" in sections:
            env_ext: list[str] = list()
            script.extend(self._run(sections["@run"], env_ext))
            if len(env_ext) > 0:
                if "@env" in sections:
                    sections["@env"].extend(env_ext)
                else:
                    sections["@env"] = env_ext
        # @env
        if "@env" in sections:
            script.extend(self._env(sections["@env"]))
        # @label
        if "@label" in sections:
            script.extend(self._label(sections["@label"]))
        # @entry
        if "@entry" in sections:
            if len(sections["@entry"]) != 1:
                raise TemplateSyntaxError("You can only have one entrypoint!")
            script.extend(self._entry(sections["@entry"]))
        # @post
        if "@post" in sections:
            script.extend(self._literal_section(sections["@post"]))

        return script

    def is_available(self) -> bool:
        """Check if the current system has the requested backend."""
        if shutil_which(self.executable) is None:
            return False
        return True

    @abstractmethod
    def _from(self, contents: list[str]) -> list[str]:
        """Handle the @from section."""

    @abstractmethod
    def _arguments(self, all_contents: dict[str, list[str]]) -> list[str]:
        """Handle arguments."""

    @abstractmethod
    def _copy(self, contents: list[str]) -> list[str]:
        """Handle the @copy section."""

    @abstractmethod
    def _run(self, contents: list[str], label_contents: list[str]) -> list[str]:
        """Handle the @run section. If any !enver directives are found, return a list for the @env section."""

    @abstractmethod
    def _env(self, contents: list[str]) -> list[str]:
        """Handle the @env section."""

    @abstractmethod
    def _label(self, contents: list[str]) -> list[str]:
        """Handle the @label section."""

    @abstractmethod
    def _entry(self, contents: list[str]) -> list[str]:
        """Handle the @entry section."""

    @classmethod
    def _literal_section(cls, contents: list[str]) -> list[str]:
        """Handle literal sections."""
        ret: list = [""]
        for ln in contents:
            ret.append(ln.lstrip("|"))
        return ret

    @abstractmethod
    def generate_build_cmd(self, src: str, dest: str, args: list[str] = None) -> str:
        """Generate CLI command to build."""

    @abstractmethod
    def format_image_name(self, path: Path, tag: str) -> str:
        """Create a name for the image build."""

    @abstractmethod
    def clean_up_old_image_tag(self, name: str) -> str:
        """Generate CLI command to clean up an old image."""

    @abstractmethod
    def build_exists(self, name: str) -> bool:
        """Check if an images has been built."""

    @abstractmethod
    def generate_final_image_cmd(self, src: str, dest: str) -> str:
        """Generate command to move the last image in the build to its final destination."""


class Podman(Backend):
    """Podman backend."""

    def __init__(self):
        super().__init__(name="podman", executable="podman")

    def _from(self, contents: list[str]) -> list[str]:
        return [f"FROM {contents[0]}"]

    def _arguments(self, all_contents: dict[str, list[str]]) -> list[str]:
        ret: list[str] = list()
        for si in all_contents.keys():
            for li in range(len(all_contents[si])):
                res: re_Match[str] = re_match(r".*(@@\s*(\S+)\s*@@).*", all_contents[si][li])
                if res is not None:
                    if len(res.group(2).split()) != 1:
                        raise TemplateSyntaxError("Arguments cannot have spaces in their names!")
                    else:
                        ret.append("ARG {}".format(res.group(2)))
                        all_contents[si][li] = _substitute(
                            all_contents[si][li],
                            {res.group(2): "${}".format(res.group(2))},
                            r"@@\s*(\S+)\s*@@",
                        )
        if len(ret) > 0:
            ret.insert(0, "")
        return ret

    def _copy(self, contents: list[str]) -> list[str]:
        ret: list[str] = [""]
        for ln in contents:
            if len(ln.split()) != 2:
                raise TemplateSyntaxError("Entries in @copy can only have one source and destination!", ln)
            ret.append(f"COPY {ln}")
        return ret

    def _run(self, contents: list[str], label_contents: list[str]) -> list[str]:
        ret: list[str] = [""]
        for cmd in contents:
            # process !envar directives
            alt_cmd = cmd
            if re_match(r"^!envar\s+.*", cmd):
                res = re_match(r"^!envar\s+(?P<name>\S+)\s+(?P<value>.*)$", cmd)
                alt_cmd = 'export {name}="{value}"'.format(**res.groupdict())
                label_contents.append("{name} {value}".format(**res.groupdict()))
            # generate line
            ln = ""
            # place RUN on the first line
            if cmd == contents[0]:
                ln += "RUN "
            else:
                # indent following lines
                ln += "    "
            ln += alt_cmd
            # add '&& \\' to all but the last line
            if cmd != contents[-1] and cmd[-1] != '\\': # ignore line that end in an escape
                ln += " && \\"
            ret.append(ln)
        return ret

    def _env(self, contents: list[str]) -> list[str]:
        ret: list[str] = [""]
        for env in contents:
            parts = env.split()
            # generate line
            ln = ""
            # add ENV to the first line
            if env == contents[0]:
                ln += "ENV "
            else:
                # indent following lines
                ln += "    "
            ln += f"{parts[0]}=\"{env.lstrip(parts[0]).strip(' ')}\""
            # add '\\' to all but the last line
            if env != contents[-1]:
                ln += " \\"
            ret.append(ln)
        return ret

    def _label(self, contents: list[str]) -> list[str]:
        ret: list[str] = [""]
        for label in contents:
            parts = label.split()
            if len(parts) != 2:
                raise TemplateSyntaxError("Label '{}' must have two parts!".format(label))
            # generate line
            ln = ""
            # add LABEL to the first line
            if label == contents[0]:
                ln += "LABEL "
            else:
                # indent following lines
                ln += "    "
            ln += f"{parts[0]}=\"{label.lstrip(parts[0]).strip(' ')}\""
            # add '\\' to all but the last line
            if label != contents[-1]:
                ln += " \\"
            ret.append(ln)
        return ret

    def _entry(self, contents: list[str]) -> list[str]:
        return ["", "ENTRYPOINT {}".format(contents[0].split())]

    def generate_build_cmd(self, src: str, dest: str, args: list = None) -> str:
        cmd: list[str] = ["{} build".format(self.executable)]
        # arguments
        if args is not None and len(args) > 0:
            cmd.append(" ".join(_ for _ in args) if args is not None else "")
        # script
        cmd.append("-f {}".format(src))
        # destination
        cmd.append("-t {}".format(dest))
        # build dir
        cmd.append(".")
        return " ".join(_ for _ in cmd) + ";"

    def format_image_name(self, path: Path, tag: str) -> str:
        return "{}{}{}".format("localhost/" if "/" not in tag else "", tag, ":latest" if ":" not in tag else "")

    def clean_up_old_image_tag(self, name: str) -> str:
        return "podman rmi {}".format(name)

    def build_exists(self, name: str) -> bool:
        return False

    def generate_final_image_cmd(self, src: str, dest: str) -> str:
        return "{} tag {} {}".format(self.executable, src, dest)


class Apptainer(Backend):
    """Apptainer backend."""

    def __init__(self):
        super().__init__(name="apptainer", executable="apptainer")

    def _from(self, contents: list[str]) -> list[str]:
        ret: list[str] = list()
        if re_match(r"^.*\.sif$", contents[0]):
            ret.append("Bootstrap: localimage")
        elif re_match(r"^.*\/.*:.*$", contents[0]):
            ret.append("Bootstrap: docker")
        else:
            raise TemplateSyntaxError("Unknown source format in @from!", contents[0])
        ret.append("From: {}".format(contents[0]))
        return ret

    def _arguments(self, all_contents: dict[str, list[str]]) -> list[str]:
        for si in all_contents.keys():
            for li in range(len(all_contents[si])):
                res = re_match(r".*(@@\s*(\S+)\s*@@).*", all_contents[si][li])
                if res is not None:
                    if len(res.group(2).split()) != 1:
                        raise TemplateSyntaxError("Arguments cannot have spaces in their names!")
                    else:
                        all_contents[si][li] = _substitute(
                            all_contents[si][li],
                            {res.group(2): f"{{{{ {res.group(2)} }}}}"},
                            r"@@\s*(\S+)\s*@@",
                        )
        return list()

    def _copy(self, contents: list[str]) -> list[str]:
        ret: list[str] = ["", "%files"]
        for ln in contents:
            if len(ln.split()) != 2:
                raise TemplateSyntaxError("Your '@copy' can only have one source and destination!", ln)
            ret.append("{}".format(ln))
        return ret

    def _run(self, contents: list[str], label_contents: list[str]) -> list[str]:
        ret: list[str] = ["", "%post"]
        for cmd in contents:
            # handel !envar directives
            if re_match(r"^!envar\s+.*", cmd):
                res = re_match(r"^!envar\s+(?P<name>\S+)\s+(?P<value>.*)$", cmd)
                cmd = 'export {name}="{value}"'.format(**res.groupdict())
                label_contents.append("{name} {value}".format(**res.groupdict()))
            ret.append("{}".format(cmd))
        return ret

    def _env(self, contents: list[str]) -> list[str]:
        ret: list[str] = ["", "%environment"]
        for env in contents:
            parts = env.split()
            ret.append('export {}="{}"'.format(parts[0], env.lstrip(parts[0]).strip(" ")))
        return ret

    def _label(self, contents: list[str]) -> list[str]:
        ret: list[str] = ["", "%labels"]
        for label in contents:
            parts = label.split()
            ret.append("{} {}".format(parts[0], label.lstrip(parts[0]).strip(" ")))
        return ret

    def _entry(self, contents: list[str]) -> list[str]:
        return ["", "%runscript", "{}".format(contents[0])]

    def generate_build_cmd(self, src: str, dest: str, args: list = None) -> str:
        cmd: list[str] = ["{} build".format(self.executable)]
        # arguments
        if args is not None and len(args) > 0:
            cmd.append(" ".join(_ for _ in args) if args is not None else "")
        # destination
        cmd.append("{}".format(dest))
        # script
        cmd.append("{}".format(src))
        return " ".join(_ for _ in cmd) + ";"

    def format_image_name(self, path: Path, tag: str) -> str:
        return "{}{}".format(Path.joinpath(path, tag), ".sif" if ".sif" not in tag else "")

    def clean_up_old_image_tag(self, name: str) -> str:
        return "echo"

    def build_exists(self, name: str) -> bool:
        if Path(name).is_file():
            return True
        return False

    def generate_final_image_cmd(self, src: str, dest: str) -> str:
        return "cp {} {}".format(src, dest)


def get_backend() -> Backend:
    backend = config.get("velocity:backend")
    if backend == "podman":
        b = Podman()
    elif backend == "apptainer":
        b = Apptainer()
    else:
        raise BackendNotSupported(backend)

    # check that backend is available
    if b.is_available():
        return b
    else:
        raise BackendNotAvailable("Your system does not have the '{}' backend".format(b.name))



"""Build velocity images."""

import datetime
import shutil
import os
from timeit import default_timer as timer
from subprocess import Popen, PIPE
from queue import SimpleQueue
from threading import Thread
from pathlib import Path
from colorama import Fore, Style
from platform import processor as arch
from ._config import config
from ._graph import Image
from ._print import header_print, indent_print, TextBlock
from ._backends import get_backend, Backend


def read_pipe(pipe: PIPE, topic: SimpleQueue, prefix: str, log: SimpleQueue) -> None:
    """Read a subprocess PIPE and place lines on topic queue and log queue."""
    while True:
        ln = pipe.readline()
        if ln == "":
            break
        else:
            topic.put(ln.strip("\n"))
            log.put("{} {}".format(prefix, ln.strip("\n")))


def run(cmd: str, log_file: Path = None, verbose: bool = False) -> None:
    """Run a system command logging all output to a file and print if verbose."""
    # open log file (set to False if none is provided)
    file = open(log_file, "w") if log_file is not None else False

    log = SimpleQueue()
    stdout = SimpleQueue()
    stderr = SimpleQueue()

    process = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE, universal_newlines=True)

    out = Thread(target=read_pipe, args=(process.stdout, stdout, "STDOUT:", log))
    err = Thread(target=read_pipe, args=(process.stderr, stderr, "STDERR:", log))

    out.start()
    err.start()

    # loop for real time output
    while process.poll() is None:
        if verbose and not stdout.empty():
            indent_print([TextBlock(stdout.get(), fore=Fore.GREEN, style=Style.DIM)])
        if file and not log.empty():
            file.write(log.get() + "\n")
            file.flush()  # TODO is this needed?

    out.join()
    err.join()

    # clear stdout & log
    if verbose:
        while not stdout.empty():
            indent_print([TextBlock(stdout.get(), fore=Fore.GREEN, style=Style.DIM)])

    if file:
        while not log.empty():
            file.write(log.get() + "\n")
            file.flush()  # TODO is this needed?
        file.close()

    # if an error was encountered exit with the subprocess exit code
    if process.poll() != 0:
        while stderr.qsize():
            indent_print([TextBlock(stderr.get(), fore=Fore.RED, style=Style.DIM)])
        exit(process.poll())


class ImageBuilder:
    """Image building class."""

    def __init__(
        self,
        bt: tuple[Image],
        build_name: str = None,
        dry_run: bool = False,
        remove_tags: bool = True,
        clean_build_dir: bool = False,
        verbose: bool = False,
    ) -> None:
        self.build_units: tuple[Image] = bt
        self.build_name: str = build_name
        self.dry_run: bool = dry_run
        self.remove_tags: bool = remove_tags
        self.clean_build_dir: bool = clean_build_dir
        self.verbose: bool = verbose

        self.backend_engine: Backend = get_backend()

        # create build_dir if it does not exist
        self.build_dir = Path(config.get("velocity:build_dir"))
        self.build_dir.mkdir(mode=0o777, parents=True, exist_ok=True)

        self.variables: dict[str, str] = dict()
        for i in self.build_units:
            self.variables["__{}__version__".format(i.name)] = i.version.__str__()
            self.variables["__{}__version_major__".format(i.name)] = i.version.major.__str__()
            self.variables["__{}__version_minor__".format(i.name)] = i.version.minor.__str__()
            self.variables["__{}__version_patch__".format(i.name)] = i.version.patch.__str__()
            self.variables["__{}__version_suffix__".format(i.name)] = i.version.suffix.__str__()

    def build(self) -> None:
        """Launch image builds."""
        # store pwd
        pwd = Path(Path().absolute())

        # clean build_dir
        if self.clean_build_dir:
            for entry in self.build_dir.iterdir():
                if entry.is_dir():
                    shutil.rmtree(entry)
                else:
                    entry.unlink()

        last = str()  # last image that was built
        build_names: list[str] = list()
        for u in self.build_units:
            name = self.backend_engine.format_image_name(
                Path.joinpath(self.build_dir, "{}-{}-{}".format(u.name, u.version, u.id)), u.id
            )
            self._build_image(u, last, name)
            last = name
            build_names.append(name)

        tag = str(
            self.build_name
            if self.build_name is not None
            else "{}__{}-{}".format(
                "_".join(f"{bu.name}-{bu.version}" for bu in reversed(self.build_units)),
                config.get("velocity:system"),
                config.get("velocity:distro"),
            )
        )

        final_name = self.backend_engine.format_image_name(Path(pwd.absolute()), tag)
        if not self.dry_run:
            run(self.backend_engine.generate_final_image_cmd(last, final_name))
        header_print([TextBlock("BUILT: "), TextBlock(final_name, fore=Fore.MAGENTA, style=Style.BRIGHT)])

        if not self.dry_run and self.remove_tags:
            for bn in build_names:
                run(self.backend_engine.clean_up_old_image_tag(bn))

        # go back to the starting dir
        os.chdir(pwd)

    def _build_image(self, unit: Image, src_image: str, name: str):
        """Build an individual image."""
        # print start of build
        header_print(
            [
                TextBlock(unit.id, fore=Fore.RED, style=Style.BRIGHT),
                TextBlock(": BUILD "),
                TextBlock("{}@{}".format(unit.name, unit.version), fore=Fore.MAGENTA, style=Style.BRIGHT),
                TextBlock("{} ...".format(" --DRY-RUN" if self.dry_run else "")),
            ]
        )

        start = timer()

        # create build dir and go to it
        build_sub_dir = Path.joinpath(self.build_dir, "{}-{}-{}".format(unit.name, unit.version, unit.id))
        build_sub_dir.mkdir(mode=0o744, exist_ok=True)
        os.chdir(build_sub_dir)

        # copy additional files
        if len(unit.files) > 0:
            header_print([TextBlock(unit.id, fore=Fore.RED, style=Style.BRIGHT), TextBlock(": COPYING FILES ...")])
            for entry in unit.files:
                # print copy operation
                if self.verbose:
                    indent_print(
                        [
                            TextBlock(
                                "DIR: " if Path.joinpath(unit.path, "files", entry).is_dir() else "FILE: ",
                                fore=Fore.YELLOW,
                                style=Style.BRIGHT,
                            ),
                            TextBlock(str(Path.joinpath(unit.path, "files", entry).absolute()), fore=Fore.GREEN),
                            TextBlock(" -> ", fore=Fore.YELLOW, style=Style.BRIGHT),
                            TextBlock(str(Path.joinpath(build_sub_dir, entry).absolute()), fore=Fore.GREEN),
                        ]
                    )
                if Path.joinpath(unit.path, "files", entry).is_dir():
                    shutil.copytree(Path.joinpath(unit.path, "files", entry), Path.joinpath(build_sub_dir, entry))
                else:
                    shutil.copy(Path.joinpath(unit.path, "files", entry), Path.joinpath(build_sub_dir, entry))

        # parse template and create script...
        header_print([TextBlock(unit.id, fore=Fore.RED, style=Style.BRIGHT), TextBlock(": GENERATING SCRIPT ...")])
        if self.verbose:
            indent_print(
                [
                    TextBlock("SCRIPT: ", fore=Fore.YELLOW, style=Style.BRIGHT),
                    TextBlock(str(Path.joinpath(build_sub_dir, "script")), fore=Fore.GREEN),
                ]
            )

        # get and update script variables
        script_variables = unit.variables.copy()
        script_variables.update({"__name__": unit.name})
        script_variables.update({"__version__": str(unit.version)})
        script_variables.update({"__version_major__": str(unit.version.major)})
        script_variables.update({"__version_minor__": str(unit.version.minor)})
        script_variables.update({"__version_patch__": str(unit.version.patch)})
        script_variables.update({"__version_suffix__": str(unit.version.suffix)})
        script_variables.update({"__timestamp__": str(datetime.datetime.now())})
        script_variables.update(
            {"__threads__": str(int(os.cpu_count() * 0.75) if int(os.cpu_count() * 0.75) < 16 else 16)}
        )
        script_variables.update({"__arch__": arch()})
        if src_image is not None:
            script_variables.update({"__base__": src_image})
        script_variables.update(self.variables)

        script = self.backend_engine.generate_script(unit, script_variables)
        # write out script
        with open(Path.joinpath(build_sub_dir, "script"), "w") as out_file:
            for line in script:
                if self.verbose:
                    indent_print([TextBlock(line, fore=Fore.BLUE, style=Style.DIM)])
                out_file.writelines(line + "\n")

        # run prolog & build
        build_cmd = self.backend_engine.generate_build_cmd(
            str(Path.joinpath(build_sub_dir, "script")), name, unit.arguments
        )
        build_file_path = Path.joinpath(build_sub_dir, "build")
        build_contents = ["#!/usr/bin/env bash"]

        if unit.prolog is not None:
            header_print(
                [
                    TextBlock(unit.id, fore=Fore.RED, style=Style.BRIGHT),
                    TextBlock(": RUNNING PROLOG ... && BUILDING ..."),
                ]
            )
            build_contents.append(unit.prolog.strip("\n"))
            build_contents.append(build_cmd)

        else:
            header_print([TextBlock(unit.id, fore=Fore.RED, style=Style.BRIGHT), TextBlock(": BUILDING ...")])
            build_contents.append(build_cmd)

        with open(build_file_path, "w") as build_file:
            for line in build_contents:
                build_file.write(line + "\n")
                if self.verbose:
                    indent_print([TextBlock(line, fore=Fore.YELLOW, style=Style.BRIGHT)])

        build_file_path.chmod(0o744)

        if not self.dry_run:
            if self.backend_engine.build_exists(name):
                if self.verbose:
                    indent_print(
                        [TextBlock("Using cached image {} ...".format(name), fore=Fore.GREEN, style=Style.DIM)]
                    )
            else:
                run(str(build_file_path.absolute()), log_file=Path.joinpath(build_sub_dir, "log"), verbose=self.verbose)

        end = timer()

        header_print(
            [
                TextBlock(unit.id, fore=Fore.RED, style=Style.BRIGHT),
                TextBlock(": IMAGE "),
                TextBlock(name, fore=Fore.MAGENTA, style=Style.BRIGHT),
                TextBlock(" ("),
                TextBlock("{}@{}".format(unit.name, unit.version), fore=Fore.MAGENTA, style=Style.BRIGHT),
                TextBlock(") BUILT ["),
                TextBlock(str(datetime.timedelta(seconds=round(end - start))), fore=Fore.MAGENTA, style=Style.BRIGHT),
                TextBlock("]"),
            ]
        )
        print()  # new line



"""Printing formatter."""

from colorama import Fore, Back, Style


class TextBlock:
    """Defines a block of text and its styling."""

    def __init__(
        self,
        text: str,
        fore: Fore = Fore.RESET,
        back: Back = Back.RESET,
        style: Style = Style.NORMAL,
    ) -> None:
        self.text = text
        self.fore = fore
        self.back = back
        self.style = style


def bare_print(tb: list[TextBlock]) -> None:
    """Print a list of TextBlocks."""
    for text_block in tb:
        print(
            "{}{}{}{}{}".format(
                text_block.fore,
                text_block.back,
                text_block.style,
                text_block.text,
                Style.RESET_ALL,
            ),
            end="",
        )
    # print a newline at the end
    print()


def header_print(tb: list[TextBlock]) -> None:
    """Print a list of TextBlocks with a preceding '==> ' block."""
    text_blocks = [TextBlock("==> ", fore=Fore.GREEN, style=Style.BRIGHT)]
    text_blocks.extend(tb)
    bare_print(text_blocks)


def indent_print(tb: list[TextBlock]) -> None:
    """Print a list of TextBlocks with a preceding 4 spaces."""
    text_blocks = [TextBlock("    ", fore=Fore.GREEN, style=Style.BRIGHT)]
    text_blocks.extend(tb)
    bare_print(text_blocks)
